rm(list=ls())
require(text2vec)
require(keras)
require(PerfMeas)

outcomes <- c("rds", "ivh", "nec", "rop", "bpd", "pda", "pvl", "sepsis", "pulm_hem", "cp", "pulm_htn", "hyperbilirubinemia", "death", "mas", "atelectasis", "candidiasis",    
              "cardiac_failure", "cardio_instanbility", "other_cns", "gastroesophageal", "rf", "polycythemia", "seizures", "anemia")

# generate embeddings
embedding_dim <- 128
max_concepts <- 10000                                                   # We will only consider the top 10000 concepts in the dataset
set.seed(987); embeddings <- data.frame(matrix(rnorm(max_concepts*embedding_dim), ncol = embedding_dim))
row.names(embeddings) <- as.character(1:max_concepts)

# generate input (i) as sequence of codes (these are randomly generated for the sole purpose of testing the code)
input_i <- data.frame()
for (nn in 1:30000) {
  set.seed(123); input_i[nn,"sequence"] <- paste0(floor(runif(n = rpois(1, lambda = 20), min = 1, max = 15000)), collapse=" ")
}

# generate input (ii)  (these are randomly generated for the sole purpose of testing the code)
set.seed(321); input_ii <- data.frame(matrix(rnorm(30000*20), ncol = 20))

# generate outcome data  (these are randomly generated for the sole purpose of testing the code)
set.seed(456); y <- data.frame(matrix(rbinom(30000*24, size=1, prob=0.1), ncol=24))
names(y) <- outcomes

tokenizer <- text_tokenizer(num_words = max_concepts) %>% 
  fit_text_tokenizer(c(input_i[,"sequence"]))

sequences <- texts_to_sequences(tokenizer, c(input_i[,"sequence"]))
concept_index = tokenizer$word_index

maxlen <- 1000                                                          # Cut sequences after 1000 concepts 
data_pad <- pad_sequences(sequences, maxlen = maxlen, truncating = 'pre', padding = 'pre')

embeddings_index <- new.env(hash = TRUE, parent = emptyenv())
for (i in 1:max_concepts) {
  concept <- rownames(embeddings)[i]
  embeddings_index[[concept]] <- as.double(embeddings[i,])
}

embedding_matrix <- array(0, c(max_concepts, embedding_dim))
for (concept in names(concept_index)) {
  index <- concept_index[[concept]]
  if (index < max_concepts) {
    embedding_vector <- embeddings_index[[concept]]
    if (!is.null(embedding_vector))
      # Concept not found in the embedding index will be all zeros.
      embedding_matrix[index+1,] <- embedding_vector
  }
}

build_model <- function() {
  
  lstm_input <- layer_input(shape = maxlen, name = 'lstm_input')
  
  lstm_out <- lstm_input %>%
    layer_embedding(input_dim = max_concepts, output_dim = embedding_dim, input_length = maxlen, name='embed', mask_zero = T) %>% 
    bidirectional(layer_lstm(units=128, dropout = 0.2, recurrent_dropout = 0.2, activation="tanh", recurrent_activation="sigmoid")) %>%
    layer_dropout(rate=0.2) 
  
  fixed_input <- layer_input(shape = ncol(input_ii), name = 'fixed_input') 
  
  fixed_out <- fixed_input %>%
    layer_batch_normalization() %>%
    layer_dense(units = 4, activation = "tanh") %>%
    layer_dropout(rate=0.2)
  
  core <- layer_concatenate(c(lstm_out, fixed_out)) %>%
    layer_batch_normalization() %>%
    layer_dense(units = 64, activation = 'tanh') %>%
    layer_dropout(rate=0.2)
  
  rds <- core %>%
    layer_dense(units = 32, activation = 'relu') %>%
    layer_dropout(rate=0.2) %>%
    layer_dense(units = 16, activation = 'tanh') %>%
    layer_dropout(rate=0.2) %>%
    layer_dense(units = 1, activation = 'sigmoid', name = 'rds')
  
  ivh <- core %>%
    layer_dense(units = 32, activation = 'relu') %>%
    layer_dropout(rate=0.2) %>%
    layer_dense(units = 16, activation = 'tanh') %>%
    layer_dropout(rate=0.2) %>%
    layer_dense(units = 1, activation = 'sigmoid', name = 'ivh')
  
  nec <- core %>%
    layer_dense(units = 32, activation = 'relu') %>%
    layer_dropout(rate=0.25) %>%
    layer_dense(units = 16, activation = 'tanh') %>%
    layer_dropout(rate=0.25) %>%
    layer_dense(units = 1, activation = 'sigmoid', name = 'nec')
  
  rop <- core %>%
    layer_dense(units = 32, activation = 'relu') %>%
    layer_dropout(rate=0.25) %>%
    layer_dense(units = 16, activation = 'tanh') %>%
    layer_dropout(rate=0.25) %>%
    layer_dense(units = 1, activation = 'sigmoid', name = 'rop')
  
  bpd <- core %>%
    layer_dense(units = 32, activation = 'relu') %>%
    layer_dropout(rate=0.2) %>%
    layer_dense(units = 16, activation = 'tanh') %>%
    layer_dropout(rate=0.2) %>%
    layer_dense(units = 1, activation = 'sigmoid', name = 'bpd')
  
  pda <- core %>%
    layer_dense(units = 32, activation = 'relu') %>%
    layer_dropout(rate=0.2) %>%
    layer_dense(units = 16, activation = 'tanh') %>%
    layer_dropout(rate=0.2) %>%
    layer_dense(units = 1, activation = 'sigmoid', name = 'pda')
  
  pvl <- core %>%
    layer_dense(units = 32, activation = 'relu') %>%
    layer_dropout(rate=0.2) %>%
    layer_dense(units = 16, activation = 'tanh') %>%
    layer_dropout(rate=0.2) %>%
    layer_dense(units = 1, activation = 'sigmoid', name = 'pvl')
  
  sepsis <- core %>%
    layer_dense(units = 32, activation = 'relu') %>%
    layer_dropout(rate=0.2) %>%
    layer_dense(units = 16, activation = 'tanh') %>%
    layer_dropout(rate=0.2) %>%
    layer_dense(units = 1, activation = 'sigmoid', name = 'sepsis')
  
  pulm_hem <- core %>%
    layer_dense(units = 32, activation = 'relu') %>%
    layer_dropout(rate=0.25) %>%
    layer_dense(units = 16, activation = 'tanh') %>%
    layer_dropout(rate=0.25) %>%
    layer_dense(units = 1, activation = 'sigmoid', name = 'pulm_hem')
  
  cp <- core %>%
    layer_dense(units = 32, activation = 'relu') %>%
    layer_dropout(rate=0.25) %>%
    layer_dense(units = 16, activation = 'tanh') %>%
    layer_dropout(rate=0.25) %>%
    layer_dense(units = 1, activation = 'sigmoid', name = 'cp')
  
  pulm_htn <- core %>%
    layer_dense(units = 32, activation = 'relu') %>%
    layer_dropout(rate=0.25) %>%
    layer_dense(units = 16, activation = 'tanh') %>%
    layer_dropout(rate=0.25) %>%
    layer_dense(units = 1, activation = 'sigmoid', name = 'pulm_htn')
  
  hyperbilirubinemia <- core %>%
    layer_dense(units = 32, activation = 'relu') %>%
    layer_dropout(rate=0.2) %>%
    layer_dense(units = 16, activation = 'tanh') %>%
    layer_dropout(rate=0.2) %>%
    layer_dense(units = 1, activation = 'sigmoid', name = 'hyperbilirubinemia')
  
  death <- core %>%
    layer_dense(units = 32, activation = 'relu') %>%
    layer_dropout(rate=0.2) %>%
    layer_dense(units = 16, activation = 'tanh') %>%
    layer_dropout(rate=0.25) %>%
    layer_dense(units = 1, activation = 'sigmoid', name = 'death')
  
  mas <- core %>%
    layer_dense(units = 32, activation = 'relu') %>%
    layer_dropout(rate=0.2) %>%
    layer_dense(units = 16, activation = 'tanh') %>%
    layer_dropout(rate=0.2) %>%
    layer_dense(units = 1, activation = 'sigmoid', name = 'mas')
  
  atelectasis <- core %>%
    layer_dense(units = 32, activation = 'relu') %>%
    layer_dropout(rate=0.2) %>%
    layer_dense(units = 16, activation = 'tanh') %>%
    layer_dropout(rate=0.2) %>%
    layer_dense(units = 1, activation = 'sigmoid', name = 'atelectasis')
  
  candidiasis <- core %>%
    layer_dense(units = 32, activation = 'relu') %>%
    layer_dropout(rate=0.2) %>%
    layer_dense(units = 16, activation = 'tanh') %>%
    layer_dropout(rate=0.2) %>%
    layer_dense(units = 1, activation = 'sigmoid', name = 'candidiasis')
  
  cardiac_failure <- core %>%
    layer_dense(units = 32, activation = 'relu') %>%
    layer_dropout(rate=0.2) %>%
    layer_dense(units = 16, activation = 'tanh') %>%
    layer_dropout(rate=0.2) %>%
    layer_dense(units = 1, activation = 'sigmoid', name = 'cardiac_failure')
  
  cardio_instanbility <- core %>%
    layer_dense(units = 32, activation = 'relu') %>%
    layer_dropout(rate=0.2) %>%
    layer_dense(units = 16, activation = 'tanh') %>%
    layer_dropout(rate=0.2) %>%
    layer_dense(units = 1, activation = 'sigmoid', name = 'cardio_instanbility')
  
  other_cns <- core %>%
    layer_dense(units = 32, activation = 'relu') %>%
    layer_dropout(rate=0.2) %>%
    layer_dense(units = 16, activation = 'tanh') %>%
    layer_dropout(rate=0.2) %>%
    layer_dense(units = 16, activation = 'tanh') %>%
    layer_dropout(rate=0.2) %>%
    layer_dense(units = 1, activation = 'sigmoid', name = 'other_cns')
  
  gastroesophageal <- core %>%
    layer_dense(units = 32, activation = 'relu') %>%
    layer_dropout(rate=0.2) %>%
    layer_dense(units = 16, activation = 'tanh') %>%
    layer_dropout(rate=0.2) %>%
    layer_dense(units = 1, activation = 'sigmoid', name = 'gastroesophageal')
  
  rf <- core %>%
    layer_dense(units = 32, activation = 'relu') %>%
    layer_dropout(rate=0.2) %>%
    layer_dense(units = 16, activation = 'tanh') %>%
    layer_dropout(rate=0.2) %>%
    layer_dense(units = 1, activation = 'sigmoid', name = 'rf')
  
  polycythemia <- core %>%
    layer_dense(units = 32, activation = 'relu') %>%
    layer_dropout(rate=0.2) %>%
    layer_dense(units = 16, activation = 'tanh') %>%
    layer_dropout(rate=0.2) %>%
    layer_dense(units = 1, activation = 'sigmoid', name = 'polycythemia')
  
  seizures <- core %>%
    layer_dense(units = 32, activation = 'relu') %>%
    layer_dropout(rate=0.2) %>%
    layer_dense(units = 16, activation = 'tanh') %>%
    layer_dropout(rate=0.2) %>%
    layer_dense(units = 1, activation = 'sigmoid', name = 'seizures')
  
  anemia <- core %>%
    layer_dense(units = 32, activation = 'relu') %>%
    layer_dropout(rate=0.2) %>%
    layer_dense(units = 16, activation = 'tanh') %>%
    layer_dropout(rate=0.2) %>%
    layer_dense(units = 1, activation = 'sigmoid', name = 'anemia')
  
  model <- keras_model(
    inputs = list(lstm_input, fixed_input),
    outputs = list(rds,ivh,nec,rop,bpd,pda,pvl,sepsis,pulm_hem,cp,pulm_htn,hyperbilirubinemia,death,mas,atelectasis,candidiasis,cardiac_failure,
                   cardio_instanbility,other_cns,gastroesophageal,rf,polycythemia,seizures,anemia))
  
  get_layer(model, name='embed') %>% 
    set_weights(list(embedding_matrix)) %>% 
    freeze_weights()
  
  model %>% compile(
    optimizer = optimizer_adam(lr = 0.01, beta_1 = 0.9,beta_2 = 0.999,epsilon = NULL,decay = 0.01,amsgrad = FALSE, clipnorm = NULL, clipvalue = NULL),
    loss = "binary_crossentropy",
    metrics = c("acc"))
  
}


k <- 5                      # number of folds
set.seed(1131)
indices <- sample(1:nrow(data_pad))
folds <- cut(indices, breaks = k, labels = FALSE)
pred_scores <- data.frame()

for (i in 1:k) {
  cat("processing fold #", i, "\n")
  test_indices <- which(folds == i, arr.ind = TRUE)
  x_test <- data_pad[test_indices,]
  x_test_f <- as.matrix(input_ii[test_indices,])
  y_test <- y[test_indices,]
  
  x_train <- data_pad[-test_indices,]
  x_train_f <- as.matrix(input_ii[-test_indices,])
  y_train <- y[-test_indices,]
  
  model <- build_model() 
  
  model %>% keras::fit(
    x = list(lstm_input = x_train, fixed_input = x_train_f),
    y = list(rds = y_train[,'rds'], ivh = y_train[,'ivh'], nec = y_train[,'nec'], rop = y_train[,'rop'], bpd = y_train[,'bpd'], pda = y_train[,'pda'],
             pvl = y_train[,'pvl'], sepsis = y_train[,'sepsis'], pulm_hem = y_train[,'pulm_hem'], cp = y_train[,'cp'], pulm_htn = y_train[,'pulm_htn'], 
             hyperbilirubinemia = y_train[,'hyperbilirubinemia'], death = y_train[,'death'], mas = y_train[,'mas'], atelectasis = y_train[,'atelectasis'], candidiasis = y_train[,'candidiasis'], 
             cardiac_failure = y_train[,'cardiac_failure'], cardio_instanbility = y_train[,'cardio_instanbility'], other_cns = y_train[,'other_cns'], gastroesophageal = y_train[,'gastroesophageal'], 
             rf = y_train[,'rf'], polycythemia = y_train[,'polycythemia'], seizures = y_train[,'seizures'], anemia = y_train[,'anemia']),
    epochs = 100,
    batch_size = 512,
    validation_split = 0.2, verbose=2,
    callbacks=list(callback_early_stopping(monitor='val_loss', verbose=1, patience=10),
                   callback_model_checkpoint(filepath='.../best_model.h5', monitor='val_loss', save_best_only=T))     # insert path to file
  )
  
  bestmodel <- load_model_hdf5('.../best_model.h5', compile = F)                                                      # insert path to file
  
  testpredictor <- bestmodel %>% predict(x=list(lstm_input = x_test, fixed_input = x_test_f))
  names(testpredictor) <- outcomes
  
  for (oo in outcomes) {
    pred_scores[test_indices,oo] <- c(testpredictor[[oo]])
  }
}

##calculate cv results
testres=data.frame()
for (i in names(pred_scores)) {
  z=precision.at.all.recall.levels(pred_scores[,i], y[,i])
  testres[which(names(pred_scores)==i),'outcome'] = i
  testres[which(names(pred_scores)==i),'PR'] = (trap.rule.integral(z$recall, z$precision))
  testres[which(names(pred_scores)==i),'AUC'] = AUC.single(pred_scores[,i], y[,i])
}
testres

rm(list=ls())
